// * ======================================================================
// * *                                                                    *
// * *  Copyright (c) 1996, University of Salzburg.                       *
// * *  All rights reserved.                                              *
// * *                                                                    *
// * ======================================================================
// * *                                                                    *  
// * *  DataCast Version 1.8                                              *
// * *                                                                    *
// * *  Author:  Manfred Hackel                                           *
// * *  File  :  Document.java                                            *
// * *  Date  :  31.01.1996                                               *
// * *                                                                    *
// * *  Description:                                                      *
// * *                                                                    *
// * *                                                                    *
// * *  Revision-History:                                                 *
// * *     1.7   Manfred Hackel   initial coding                          *
// * *     1.7R  Hilmar Linder    added RRMP support for Windows 95/NT    *
// * *     1.8   Hilmar Linder    check the OS version for RRMP support   *
// * *                            added the options -rf,-ru for RRMP      *
// * *                            added this header                       *
// * *                                                                    *
// * ======================================================================

import java.util.*;
import java.net.*;
import java.io.*;

public class Document {
	// a Document is a DataCast file (e.g. an HTML document, an inline image file), together with a description of the file

	// instance variables:
	final int 		datagramLength = 1024;		// must be the same as in all other DataCast classes
	final byte	protocolVersion = 1;

	private static Repository repository;	// link to repository of all stored files;

	private boolean subscription;		// is the document part of a service that is only availably by subscription
	private int docID = 0;			// Identification number of the document
	private long numberSegments;		// number of datagrams (segments) that are necessary to transmit this document
	private int padLength;			// # bytes at the end of data block of the last datagram that will be padded
	private long numberFiles = 0;		// number of files that belong to this service 
	private boolean isFrontpage = false;	// is this document the front page of service

	private String serviceName = new String(); 	// name of the news service to which this document belongs (e.g. "Unione Sarda") 
	private String fileName = new String(); 	// name of the file to which this document is to store (e.g. "wecome.html")
	private Vector subjects = new Vector();	// list of subjects describing the document (e.g. "Sports", "Economy" )
	private String mediaType = new String();	// media Type of the document (e.g. "Text", "Image", "Sound" )
	private Vector regions = new Vector();	// list of regions this document is referring to (e.g. "Europe", "North-America")
	private String language = new String();	// language of the document

	private boolean exists_backup = false;	// is there a pre-existing file with the same name?
	private long backupFileLength = 0;		// length of pre-existing file with the same name
	private BufferedOutputStream documentOutputStream = null;	// buffered output stream associated the file
	private StringBuffer s = new StringBuffer();  // Stringbuffer for processing the header information
	private int pos = 16;			// after init() pos gives the position of the first data byte in the datagram buffer		

//	********* Document *********+
public Document(Repository repository) {
	this.repository = repository;
}

//	********* init *********
public void init(byte[] buffer) throws Exception {
	// initialize the document, i.e. analyse the header of the segment/datagramBuffer and retrieve serviceName, subjects, categories, ...
	// method can be called only with the first datagram/segment of a document, because only the first segment has all information

	// for a description of the header of a segment see DataCastSender.java

	// check protocol version:
	if (buffer[0] != protocolVersion) {
		Exception e = new Exception("<document.init>: protocol version error");
		throw e;
	}

	// check flags 
	subscription = ((buffer[1] & 2) > 0);

	// get docID:
	docID = 0;
	for (int i = 2; i<4; i++) {
		if (buffer[i] < 0) {
			docID = (int) (docID *256 + buffer[i] +256);
		}
		else {
			docID = (int) (docID *256 + buffer[i]);
		}
	}

	// check seqNr:
	if (  (buffer[4] != 0) | (buffer[5] != 0) | (buffer[6] != 0) | ( buffer[7] != 1)) {
		// seqNr must be 1
		Exception e = new Exception("<document.init>: not first segment of a document");
		throw e;
	}

	// get numberSegments:
	numberSegments = 0;
	for (int i = 8; i<12; i++) {
		if (buffer[i] < 0) {
			numberSegments= numberSegments*256 + buffer[i] +256;
			}
		else {
			numberSegments = numberSegments*256 + buffer[i];
		}
	}

	// get padLength:
	padLength = 0;
	for (int i = 12; i<14; i++) {
		if (buffer[i] < 0) {
			padLength = padLength*256 + buffer[i] +256;
			}
		else {
			padLength= padLength*256 + buffer[i];
		}
	}

	// is this document a front page?
	if ( ( buffer[14] == 0 ) & (buffer[15] == 0)) {
		isFrontpage = false;
	}
	else {
		isFrontpage = true;
		numberFiles = 0;
		for (int i = 14; i<16; i++) {
			if (buffer[i] < 0) {
				numberFiles = numberFiles *256 + buffer[i] +256;
			}
			else {
				numberFiles = numberFiles *256 + buffer[i];
			}
		}
	}

	// get service name:
	pos = 16;
	s.setLength(0);
	while (true) {
		if (buffer[pos] != 0) {
			s.append((char) buffer[pos++]);
		}
		else {
			pos++;
			serviceName = s.toString();
			break;
		}
	}

	// get file name:
	s.setLength(0);
	while (true) {
		if (buffer[pos] != 0) {
			s.append((char) buffer[pos++]);
		}
		else {
			pos++;
			fileName = s.toString();
			break;
		}
	}

	// get description of file conent:
	boolean descriptions = true;	
	subjects.setSize(0);
	regions.setSize(0);
	while (descriptions) {
		switch (buffer[pos++]) {
			case (byte) 'S': { // subjects
				pos++; // skip "=" 
				s.setLength(0);
				while (buffer[pos] != 0) {
					s.append((char) buffer[pos++]);
				}
				pos++; // skip 0
				subjects.addElement(s.toString());
				break;
			}
			case (byte) 'T': { // categories
				pos++; // skip "=" 
				s.setLength(0);
				while (buffer[pos] != 0) {
					s.append((char) buffer[pos++]);
				}
				pos++; // skip 0
				mediaType = s.toString();
				break;
			}
			case (byte) 'R': { // regions
				pos++; // skip "=" 
				s.setLength(0);
				while (buffer[pos] != 0) {
					s.append((char) buffer[pos++]);
				}
				pos++; // skip 0
				regions.addElement(s.toString());
				break;
			}
			case (byte) 'L': { // language
				pos++; // skip "=" 
				s.setLength(0);
				while (buffer[pos] != 0) {
					s.append((char) buffer[pos++]);
				}
				pos++; // skip 0
				language = s.toString();
				break;
			}

			case 0: {
				descriptions = false; // end of descriptions
				break;
			}
			default: { // no other types of description implemented
				descriptions = false;
				Exception e = new Exception("<document.init>: corrupt document header");
				throw e;
			}
		}
	}
	// just for debugging:
/*	System.out.println("subscription: "+subscription);
	System.out.println("docID: "+docID);
	System.out.println("numberSegments: "+numberSegments);
	System.out.println("padLength: "+padLength);
	System.out.println("fileLength: "+fileLength());
	System.out.println("isFrontpage: "+isFrontpage);
	System.out.println("numberFiles: "+numberFiles);
	System.out.println("serviceName: "+serviceName);
	System.out.println("fileName: "+fileName);
	System.out.println("language: "+language);
*/
}

//	********* createNewFile *********
public void createNewFile() throws Exception{
	// check if there is enough free Disk Space
	if (repository.isEnoughFreeDiskSpace(fileLength())) {
		File file = new File(fileName);
		// if the file name contains a path, create all directories in the path
		int fromIndex = 0;
		while (fileName.indexOf("/",fromIndex) >= 0) {
			File directory = new File(fileName.substring(0,fileName.indexOf("/",fromIndex)).replace('/',file.separatorChar));
			if (!directory.exists()) {
				if (!directory.mkdir())  {
					Exception e = new Exception("directory error");
					throw e;
				}
			}
			fromIndex = fileName.indexOf("/",fromIndex)+1;
		}
		// save pre-existing file with same name
		if (file.exists()) {
			backupFileLength = file.length();
			File backupFile = new File(fileName+".bak");
			if (file.renameTo(backupFile)) {
				exists_backup = true;
			}
			else {
				exists_backup = false;
			}
		}
		else {
			backupFileLength = 0;
			exists_backup = false;
		}
		documentOutputStream = new BufferedOutputStream(new FileOutputStream(fileName),2048);
//		documentOutputStream = new BufferedOutputStream(new FileOutputStream(fileName));
	}
	else {
		Exception e = new Exception("disk full");
		throw e;
	}
}

//	********* docID *********
public int docID() {
	// returns the document ID
	return docID;
}

//	********* serviceName *********
public String serviceName() {
	// returns the service name related to this document
	return serviceName;
}

//	********* isFree *********
public boolean isFree() {
	// does this document belong to a free service or is it only available to subscribers
	return !subscription;
}

//	********* fileName *********
public String fileName() {
	// returns the file name of this document
	return fileName;
}

//	********* numberSegments *********
public long numberSegments() {
	// returns the number of datagrams (segments) that are necessary to transmit this document
	return numberSegments;
}

//	********* isFrontpage *********
public boolean isFrontpage() {
	// is this document the front page of the news service to which it belongs?
	return isFrontpage;
}

//	********* deleteFile *********
public void deleteFile() {
	// delete the file related to this document and restore backup file if it exists
	// (method will be called if an error in transmission occurs and a segment of the document is missing)
	try {
		documentOutputStream.flush();
		documentOutputStream.close();
		File file = new File(fileName);
		boolean b;
		if (file.exists()) {
			b = file.delete();
		}
		if (exists_backup) {
			File backupFile = new File(fileName+".bak");
			b = backupFile.renameTo(file);
		}
	} catch (Exception e) {}
	finally {
		if (documentOutputStream != null ) {
			try { documentOutputStream.close(); }
			catch ( IOException ioe ) {}
		}
	}
}

//	********* closeFile *********
public void closeFile () {
	// delete the backup file if it exists
	if (exists_backup) {
		File backupFile = new File(fileName+".bak");
		boolean b = backupFile.delete();
	}
	try {
 		documentOutputStream.flush();
	}
	catch(Exception e) {}
	finally {
		if (documentOutputStream != null ) {
			try { documentOutputStream.close(); }
			catch ( IOException ioe ) {}
		}
	}
}

//	********* fileLength *********
public long fileLength() {
	// returns the file length of the document's file	
	long fileLength = (datagramLength-pos)+(datagramLength-8)*(numberSegments-1)-padLength;
	return fileLength;
}

//	********* numberFiles *********
public long numberFiles() {
	// returns the number of files that belong to this service
	// value is calculates as frontpage arrives, before that the value is 0
	return numberFiles;
}

//	********* subjects *********
public Vector subjects() {
	// returns the subject list for this document
	return subjects;
}

//	********* mediaType *********
public String mediaType() {
	// returns the type of this document
	return mediaType;
}

//	********* regions *********
public Vector regions() {
	// returns the regions list for this document
	return regions;
}

//	********* language *********
public String language() {
	// returns the language of this document
	return language;
}

//	********* store *********
public void store(Segment segment) throws Exception {
//	stores the segment body to the file belonging to the document
	long seqNr = segment.seqNr();
	if ( seqNr == 1) {
		if (seqNr < numberSegments) { 
			documentOutputStream.write(segment.buffer,pos,datagramLength-pos);
		} 
		else { // last datagram
			documentOutputStream.write(segment.buffer,pos,datagramLength-pos-padLength);
			repository.decrementFreeDiskSpace(fileLength()-backupFileLength);
			repository.addFile(fileName);
		}	
	}
	else { // sequence number > 1, not the first segment in document  
		if (seqNr < numberSegments) { 
			documentOutputStream.write(segment.buffer,8,datagramLength-8);
		} 
		else { // last datagram
			documentOutputStream.write(segment.buffer,8,datagramLength-8-padLength);
			repository.decrementFreeDiskSpace(fileLength()-backupFileLength);
			repository.addFile(fileName);
		}	
	}
}

//	********* close *********
public void close() {
	try {
		closeFile();
	} catch (Exception e) {}
}

//	********* finalize *********
protected void finalize() {
	close();
}

}