/* ======================================================================
 * *                                                                    *
 * *  Copyright (c) 1996, University of Salzburg.                       *
 * *  All rights reserved.                                              *
 * *                                                                    *
 * ======================================================================
 * *                                                                    *  
 * *  RRMP Version 1.0 for WindowsNT and Linux                          *
 * *                                                                    *
 * *  Author:  Hilmar Linder, hlinder@cosy.sbg.ac.at                    *
 * *  File  :  RRMPSocket.java                                          *
 * *  Date  :  10.08.1996                                               *
 * *                                                                    *
 * *  DLL that interfaces the RRMP.DLL with the RRMPSocket class in     *
 * *  Java.                                                             *
 * *                                                                    *
 * ======================================================================
 */

/* ---- Revision note ---------------------------------------------------
 * *
 * *  $Author: hlinder $
 * *  $Date: 1995/01/26 20:40:55 $
 * *  $Source:  $
 * *  $State: Exp $
 * *
 * ----------------------------------------------------------------------
 */


//package java.net;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.SocketException;

import java.io.FileDescriptor;
import java.io.IOException;
import java.io.InterruptedIOException;

/**
 * The datagram socket class implements unreliable and limited reliabe datagrams
 * transport services.
 * @author Hilmar Linder
 */

public
class RRMPSocket {
    
 	// These native methods are implemented in RRMPJAVA.DLL. This
	// dynamic link library provides a conversion between the
	// RRMPSocket class and the RRMP.DLL
	//
  	private native void RRMPSocketCreate(int so_type);
  	private native int  RRMPSocketBind();
  	private native int  RRMPSocketConnect();
  	private native int  RRMPSocketJoin();
  	private native int  RRMPSocketLeave();
  	private native void RRMPSocketSend(byte[] buf, int len);
  	// private native int  RRMPSocketPeek(InetAddress i);
  	private native int  RRMPSocketReceive(byte[] buf, int len);
  	private native void RRMPSocketClose();
	private native void RRMPSocketTerminate();
	 
	// native methods to set RRMP specific parameters
	private native int	RRMPSocketSetRate(int rate);
	private native int	RRMPSocketSetData(int data_size);
	private native int	RRMPSocketSetFec(int fec_size);
	private native int	RRMPSocketSetMTU(int mtu_size);
	private native int	RRMPSocketSetSendBuffer(int buffer_size);
	private native int	RRMPSocketSetRecvBuffer(int buffer_size);

	// native methods to read RRMP specific parameters
	private native int	RRMPSocketGetRate();
	private native int	RRMPSocketGetData();
	private native int	RRMPSocketGetFec();
	private native int	RRMPSocketGetMTU();
	private native int	RRMPSocketGetSendBuffer();
	private native int	RRMPSocketGetRecvBuffer();

	// native method to display the internal RRMP statistics
	private native int   RRMPSocketPrintStat();

	// The receive buffer size
	private static final int RECV_BUF_SIZE = 2000;

	// The instance variables
	private String    	localInterface;
  	private int       	localPort;
	private String	 		multicastAddress;
  	private int       	multicastPort;
  	private int    	 	socket;
	private int		 		status;


   /**
    * Load net library into runtime.
    */
  static {
	  System.loadLibrary("net");
	  System.loadLibrary("rrmp");
  }


	/**
    * Creates a limited reliable RRMP socket
    */
   public RRMPSocket() throws SocketException,IOException {
		// creates a LR RRMP socket
		RRMPSocketCreate(1);
		
		// bind the RRMP socket to the local interface
      InetAddress localHost = InetAddress.getByName("localhost");
		
		// set the member variables
		localInterface = localHost.getHostName();
		localPort = 0;
		
		// bind the local interface to the socket
		status = RRMPSocketBind();
	}


   /**
    * Creates a limited reliable RRMP socket
    * @param local host to use
    */
	public RRMPSocket(String hostName, int so_type) throws SocketException,IOException {
		// creates a LR RRMP socket
		RRMPSocketCreate(so_type);

		// binds the udp socket to desired port
      InetAddress localHost = InetAddress.getByName(hostName);
	
		// set the member variables
		localInterface = hostName;
		localPort = 0;
	
		// bind the local interface to the socket
		status = RRMPSocketBind();
   }


   /**
    * Creates a limited reliable, or a unreliable RRMP socket
    * @param local host to use
    * @so_type use 0 for unreliable RRMP socket, 1 for limited reliable RRMP socket
    */
	public RRMPSocket(String hostName, int port, int so_type) throws SocketException,IOException {
		SecurityManager security = System.getSecurityManager();
		if (security != null) {
	    	security.checkListen(port);
		}
		// creates a UR or LR RRMP socket
		RRMPSocketCreate(so_type);
	
		// binds the udp socket to desired port
      InetAddress localHost = InetAddress.getByName(hostName);
	
		// set the member variables
		localInterface = hostName;
		localPort = port;
	
		// bind the local interface to the socket
		status = RRMPSocketBind();
   }


   /**
    * Sender has to connect the RRMP socket.
    */
   public synchronized void connect(String McastAddress, int port) {
		multicastAddress = McastAddress;
      multicastPort = port;
		// connect to the multicast address
		status = RRMPSocketConnect();
   }


    /**
     * Receiver has to join a multicast connection.
     */
    public synchronized void join(String McastAddress) {
		multicastAddress = McastAddress;
		// connect to the multicast address
		status = RRMPSocketJoin();
    }


    /**
     * Receiver has to leave a multicast connection before leaving.
     */
    public synchronized void leave() {
		// connect to the multicast address
		status = RRMPSocketLeave();
    }

    /**
     * Sends Datagram Packet to the destination address
     * @param DatagramPacket to be sent. The packet contains the buffer 
     * of bytes, length and destination InetAddress and port.
     * @exception IOException i/o error occurred
     */
    public void send(DatagramPacket p) throws IOException  {

	    // check the address is ok with the security manager on every send.
	    SecurityManager security = System.getSecurityManager();
	    if (security != null) {
	        security.checkConnect(p.getAddress().getHostAddress(), p.getPort());
	    }
	    RRMPSocketSend(p.getData(), p.getLength());
    }

    /**
     * Receives datagram packet.
     * @param DatagramPacket to be received.
     * On return, the DatagramPacket contains the buffer in which the 
     * data is received, packet length, sender's address and sender's port 
     * number. Blocks until some input is available.
     * @exception IOException i/o error occurred
     */
    
    public synchronized DatagramPacket receive() throws IOException {

        // check the address is ok with the security manager before every recv.
        // If there is no security mgr, just accept.
      
        SecurityManager security = System.getSecurityManager();
        if (security != null) {
	        // peek at the packet to see who it is from.
//	        InetAddress peekAddress = new InetAddress();
//	        int peekPort = RRMPSocketPeek(peekAddress);
//	        security.checkConnect(peekAddress.getHostAddress(), peekPort);
        }
	
        // read the message into a temporary buffer
	     byte[] message = new byte[RECV_BUF_SIZE];

        // If the security check succeeds, then receive the packet.
        int length = RRMPSocketReceive(message, RECV_BUF_SIZE);
        
//		  System.out.println("RRMPSocket: message length = " + length);
//      String s = new String(message, 0, 0, length);
//		  System.out.println("RRMPSocket: message        = " + s);

        DatagramPacket packet = new DatagramPacket(message, length);
        return packet;      
    }


    /**
     * Close the datagram socket.
     */
    public synchronized void close() {
	    RRMPSocketClose();
    }


    /**
     * Terminate any pending socket calls.
     */
    public synchronized void terminate() {
	    RRMPSocketTerminate();
    }


    protected synchronized void finalize() {
	    RRMPSocketClose();
    }



/*----------------------- Set / Get RRMP options ------------------------*/


    public int setRate(int rate) throws IOException  {
		 status =  RRMPSocketSetRate(rate);
		 return status;
	 }

    public int setData(int data_size) throws IOException  {
		 status = RRMPSocketSetData(data_size);
		 return status;
	 }

    public int setFec(int fec_size) throws IOException  {
		 status = RRMPSocketSetFec(fec_size);
		 return status;
	 }

    public int setMTU(int mtu_size) throws IOException  {
		 status = RRMPSocketSetMTU(mtu_size);
		 return status;
	 }

    public int setSendBuffer(int buffer_size) throws IOException  {
	    status = RRMPSocketSetSendBuffer(buffer_size);
		 return status;
	 }

    public int setRecvBuffer(int buffer_size) throws IOException  {
	    status = RRMPSocketSetRecvBuffer(buffer_size);
		 return status;
	 }

    public int getRate() throws IOException  {
		 return RRMPSocketGetRate();
	 }

    public int getData() throws IOException  {
		 return RRMPSocketGetData();
	 }

    public int getFec() throws IOException  {
		 return RRMPSocketGetFec();
	 }

    public int getMTU() throws IOException  {
		 return RRMPSocketGetMTU();
	 }

    public int getSendBuffer() throws IOException  {
	    return RRMPSocketGetSendBuffer();
	 }

    public int getRecvBuffer() throws IOException  {
	    return RRMPSocketGetRecvBuffer();
	 }

    public int printStat() throws IOException  {
	    return RRMPSocketPrintStat();
	 }


}