// * ======================================================================
// * *                                                                    *
// * *  Copyright (c) 1996, University of Salzburg.                       *
// * *  All rights reserved.                                              *
// * *                                                                    *
// * ======================================================================
// * *                                                                    *  
// * *  DataCast Version 1.8                                              *
// * *                                                                    *
// * *  Author:  Manfred Hackel                                           *
// * *  File  :  Receiver.java                                            *
// * *  Date  :  31.01.1996                                               *
// * *                                                                    *
// * *  Description:                                                      *
// * *                                                                    *
// * *                                                                    *
// * *  Revision-History:                                                 *
// * *     1.7   Manfred Hackel   initial coding                          *
// * *     1.7R  Hilmar Linder    added RRMP support for Windows 95/NT    *
// * *     1.8   Hilmar Linder    check the OS version for RRMP support   *
// * *                            added the options -rf,-ru for RRMP      *
// * *                            added this header                       *
// * *                                                                    *
// * ======================================================================

import java.util.*;
import java.net.*;
import java.io.IOException;

public class Receiver extends Thread {
/*	this is the most important class of the DataCast Receiver system; it connects to a socket and processes 
	the incoming datagrams
*/
	// instance variables
	final int	datagramLength = 1024;	// must be the same as in all other DataCast classes
	final byte	protocolVersion = 1;		// if changed, change as well in class Segment

	private UserPreferences userPreferences;
	private Repository repository;
	private Recorder recorder;
	private String transportProtocol;	// UDP or RRMP
	private DatagramSocket socket = null;
	private RRMPSocket rsocket = null;
	private Document document = null;
	private byte[] userID = new byte[4];

static {
	Properties p = System.getProperties();
	String os_name = p.getProperty("os.name");
	if (os_name.startsWith("Windows") == true) {  
	   // RRMP is only supported under Windows 95/NT
	   System.loadLibrary("rrmpjava");
	}
}

//	********** userID ***********
public Receiver(UserPreferences userPreferences, Repository repository, Recorder recorder, String transportProtocol) throws Exception{ 
	super();
	this.userPreferences = userPreferences;
	this.repository= repository;
	this.recorder = recorder;
	this.transportProtocol = transportProtocol;
	if (transportProtocol.compareTo("UDP") == 0) {
		//	System.out.println("UDP initialisation");
		try {
			int port = this.userPreferences.getPortNumber();
			//String multicastAddr = this.userPreferences.getMulticastAddr();
			//InetAddress multicastGroup = InetAddress.getByName(multicastAddr);
			InetAddress inetaddr = InetAddress.getLocalHost();
			socket = new DatagramSocket(port); // DatagramSocket(port);
			//System.out.println(" Multicast Address:   "+ multicastAddr);
			//System.out.println(" Multicast Group:       "+ multicastGroup);
			System.out.println(" Multicast Interface: "+ inetaddr.getHostName());
			System.out.println(" Port:                "+ port);
			//socket.joinGroup(multicastGroup); // multicastGroup does not work with standard java sockets
		}
		catch (Exception e) {
			recorder.receiverError(e);
		}
	}
	else {
		if (transportProtocol.startsWith("RRMP") == true) {
			// Check the OS name
			Properties p = System.getProperties();
			String os_name = p.getProperty("os.name");
			if (os_name.startsWith("Windows") == true) {  
				int port = this.userPreferences.getPortNumber();
				String multicastAddr = this.userPreferences.getMulticastAddr();
				InetAddress	inetaddr = InetAddress.getLocalHost();
				if (transportProtocol.compareTo("RRMP-fec") == 0) {
					rsocket = new RRMPSocket(inetaddr.getHostName(), port, 1);
					System.out.println(" Using the limited reliable multicast transport service");
  				} else {
					rsocket = new RRMPSocket(inetaddr.getHostName(), port, 0);
					System.out.println(" Using the unreliable multicast transport service");
  				}
    				rsocket.join(multicastAddr);
				// Set the FEC characteristics
				rsocket.setData(20);
				rsocket.setFec(10);
				System.out.println(" Multicast Address:   "+ multicastAddr);
				System.out.println(" Multicast Interface: "+ inetaddr.getHostName());
				System.out.println(" Port:                "+ port);
			}
			else {
				throw (new Exception("RRMP is only supported under Windows 95/NT"));
			}
		}
		else {
			throw (new Exception("transport protocol not supported"));
		}
	}

	// get host's userID (=IP address for the time being, later it will be the Set-Top-Box ID)
	try {
		InetAddress inetAddress = null;
		inetAddress = InetAddress.getLocalHost();
		userID = inetAddress.getAddress();
		// just for debugging:
		StringBuffer sbuffer = new StringBuffer();
		for (int i=0; i< 4;++i) {
			if (userID[i] < 0) {
				sbuffer.append(256+userID[i]);
			}
			else {
				sbuffer.append(userID[i]);
			}
			sbuffer.append('.');
		}
		System.out.println(" my user ID is "+sbuffer.toString());
	}
	catch(Exception e) {
		System.out.println("<Receiver.init> "+e.toString());
	}
}

//	********** close ***********
public void close() {
	// stops the receiver and closes the socket
	if ((socket != null ) || (rsocket != null)) {
		try {
			InetAddress inetaddr2 = InetAddress.getByName("localhost");
			byte[] buffer = new byte[datagramLength];
			// protocol version:
			buffer[0] = protocolVersion;
			// flags:
			buffer[1] = 0; // control packet / terminate receiver

			if (transportProtocol.startsWith("RRMP") == true) {
				try {
					rsocket.printStat();		
				}
				catch (Exception e) {
					System.out.println(e.toString());
				}
			} else {
				DatagramSocket socket2 = new DatagramSocket(); // DatagramSocket
				DatagramPacket dp2 = new DatagramPacket(buffer, datagramLength, inetaddr2, userPreferences.getPortNumber());
				socket2.send(dp2);
				//socket2.leaveGroup(multicastGroup);
				socket2.close();
			}
		}
		catch (Exception e) {
			System.out.println(e.toString());
		}
	}
/*	if (document != null) {
		document.close();
	}
*/
}

//	********** finalize ***********
protected void finalize() {
	close();
}

//	********** suspend ***********
public void suspend(String m) {
	recorder.writeMessage("\n"+m);
	suspend();
}

//	********** resume ***********
public void resume(String m) {
	recorder.writeMessage(m);
	resume();
}

//	********** start ***********
public void start(String m) {
	recorder.writeMessage(m);
	start();
}

//	********** run ***********
public void run() {
	long expectedSeqNr = 1;
	byte datagramBuffer[] = new byte[datagramLength];
	DatagramPacket dp = new DatagramPacket(datagramBuffer, datagramLength);
	document = new Document(repository);
	Segment segment= new Segment();

	//state: <Wait for new document> or <Wait for next segment>
	// if expectedSeqNr == 1 then state = <Wait for new document> else state = <Wait for next segment>
	try {
		// do: "listen on UDP port"

		while (true) {
			if (transportProtocol.startsWith("RRMP") == true) {
				dp = rsocket.receive();
				datagramBuffer = dp.getData();
			} else {
				socket.receive(dp);
			}
			segment.init(datagramBuffer);
			if (segment.isControlPacket()) {
				processControlPacket(datagramBuffer);
			}
			else { // process data packet

			if (expectedSeqNr == 1 ) { // new Document expected
				//state: <Datagram arrived 1> do: "check seqNr"
				if (segment.seqNr() == 1) {
					// state: <New document arriving> do: check description
					document.init(datagramBuffer);
					recorder.expectedNewDocument(document.fileName(),document.serviceName(),document.numberSegments());
					if (userPreferences.isAccepted(document)) {
						// state: <Create file> do: "create new file"
						try {
							document.createNewFile();
							// state: <Record> do: "store segment body"
							try {
								document.store(segment);
							 	if (document.numberSegments() == segment.seqNr()) { 
									// state: <Document stored>
									recorder.documentStored();
									document.closeFile(); 
									if (document.isFrontpage()) {
										// state: <Frontpage arrived>
										newService(document);
									}
									// state: <Wait for new document>
								} 
								else { // not last segment
									expectedSeqNr ++;
									// state: <Wait for next segment> 	
								}
							} catch (Exception e) {
								recorder.storageError(document.fileName(),e);
								document.deleteFile();
								// return to state: <Wait for new document>
							} 
						} catch (Exception e) {
							recorder.storageError(document.fileName(),e);
							// return to state: <Wait for new documente>
						}
	
					} 
					else { // document is not accepted
						recorder.documentRejected(document.numberSegments());
						// return to state: <Wait for new document> 
					}
				}  
				else { // not first segment of datagram
					// datagram belongs to a rejected, corrupt or truncated document
					// state: <Wait for new document>
					if (segment.docID() != document.docID()) {
						// datagram belongs to another file
						recorder.truncatedNewDocument();
					}	
					recorder.otherDatagram(segment.seqNr()); 
				}
			} 

			else { // not new document expected
				//state: <Datagram arrived 2> do: "check seqNr"
				if (segment.seqNr() == expectedSeqNr & document.docID() == segment.docID()) { // expected segment
					// state: <Record> do: "store segment body"
					try {
						recorder.expectedDatagram(segment.seqNr());
						document.store(segment);
					 	if (document.numberSegments() == segment.seqNr()) { 
							// state: Document stored>
							recorder.documentStored();
							document.closeFile();
							if (document.isFrontpage()) {
								// state: <Frontpage arrived>
								newService(document);
								// state: <Wait for new document>
							}
							expectedSeqNr = 1;
						} 
						else { // not last segment
							expectedSeqNr ++;
							// state: <Wait for next segment> 	
						}
					} catch (Exception e) {
						recorder.storageError(document.fileName(),e);
						expectedSeqNr = 1;
						document.deleteFile();
						// state: <Wait for new document>
					} 
				}
				else { // segment was not expected
					// state: <Unexpected segment> do: "delete file"
					try {
						document.deleteFile();
						if (segment.seqNr() == 1) {

							// state: New document arriving> do: check description
							document.init(datagramBuffer);
							recorder.unexpectedNewDocument(document.fileName(),document.serviceName(),document.numberSegments());
							expectedSeqNr = 1;
							if (userPreferences.isAccepted(document)) {
								// state: <Create file> do: "create new file"
								try {
									document.createNewFile();
									// state: <Store segment> do: "store segment body"
									try {
										document.store(segment);
									 	if (document.numberSegments() == segment.seqNr()) { 
											// state: Document stored>
											recorder.documentStored();
											document.closeFile();
											if (document.isFrontpage()) {
												// state: <Frontpage arrived>
												newService(document);
												expectedSeqNr = 1;
												// state: <Wait for new document>
											}
										} 
										else { // not last segment
											expectedSeqNr ++;
											// state: <Wait for next segment> 	
										}
									} catch (Exception e) {
										recorder.storageError(document.fileName(),e);
										expectedSeqNr = 1;
										document.deleteFile();
										// return to state: <Wait for new document>
									} 
								} catch (Exception e) {
									recorder.storageError(document.fileName(),e);
									expectedSeqNr = 1;
									// state: <Wait for new documente>
								}
			
							} 
							else { // document is not accepted
								recorder.documentRejected(document.numberSegments());
								expectedSeqNr = 1;
								// state: <Wait for new document> 
							}

						}
						else { // not first segment of new document
							if (document.docID() == segment.docID()) {
								recorder.segmentsMissing (document.fileName(),segment.seqNr(),false);
							}
							else {
								recorder.segmentsMissing (document.fileName(),segment.seqNr(),true);
							}
							expectedSeqNr = 1;
							// state: <Wait for new document>
						}
					} catch (Exception e) {
						recorder.writeMessage(document.fileName()+": "+e.toString());
					}
				} //  end of "segment was not expected"
				
			} // end of "not new document expected"
			if (document.docID() == document.numberFiles() & document.numberSegments() == segment.seqNr()) {
				recorder.serviceFinished();
			}
		} // end of processing of data packet
		} // end of endless loop 
 
	}
	catch (IOException e) {
		recorder.receiverError(e);
	}
	catch (Exception e) { // segment.init()
		if (e.toString().endsWith("terminate receiver")) {
			// terminate receiver
			//	System.out.println("terminate receiver message received.");
			if (transportProtocol.startsWith("RRMP") == true) {
				rsocket.leave();
				rsocket.close();
				rsocket = null;
				//System.out.println("rsocket is closed.");
			} else {
				socket.close();
				socket = null;
				//System.out.println("socket is closed.");
			}
		}
		else {
			if (e.toString().endsWith("protocol version error")) {
				recorder.protocolError(e);
			}
			else {
				System.out.println("<Receiver.run> "+e.toString());
			}
		}
	}
	finally {
		if (socket != null ) {
			socket.close();
			socket = null;
			//System.out.println("socket is finally closed.");
		}
		if (rsocket != null ) {
			try {
				rsocket.leave();
				rsocket.close();
				rsocket = null;
				//System.out.println("rsocket is finally closed.");
			}
			catch (Exception e) {
				System.out.println(e.toString());
			}
		}
	}

}

//	********** newService ***********
private void newService(Document document) {
	//sub procedure of run()
	Service service;
	Date now = new Date();
	if (document.isFree()) {
		service = new Service(document.serviceName(),document.fileName(),now);
	}
	else {
		service = new Service(document.serviceName(),document.fileName(),now,userPreferences.getAliasName(document.serviceName()));
	}	
	repository.addService(service);
	recorder.announceService(service.aliasName(),document.numberFiles());
}

//	********* processControlPacket *********
private void processControlPacket(byte[] buffer) {
	// processes a control packet (maybe it would be better to have controlPacket implemented as a class)

	// get service name:
	String serviceName = new String();
	StringBuffer s = new StringBuffer();
	s.setLength(0);
	int pos = 4;
	while (true) {
		if (buffer[pos] != 0) {
			s.append((char) buffer[pos++]);
		}
		else {
			pos++;
			serviceName = s.toString();
			break;
		}
	}

	// get service alias name:
	String serviceAlias = new String();
	s.setLength(0);
	while (true) {
		if (buffer[pos] != 0) {
			s.append((char) buffer[pos++]);
		}
		else {
			pos++;
			serviceAlias = s.toString();
			break;
		}
	}
	// just for debugging

// next two lines uncommented for debugging purposes BCN
	System.out.println("subscribed to serviceName: "+serviceName);
	System.out.println("subscribed serviceAlias: "+serviceAlias);

	// find the position of the first userID in the packet
	int userIdLength = buffer[2];
	int numberUserIDs = buffer[3];
	int posFirstUserID = 4*((pos+3)/4);
//	System.out.println("posFirstUserID: "+posFirstUserID);

	// check if userID is in list of transmitted userIDs:
	/*	later it would be wise not to exchange plain userIDs between server and client, but to apply a cryptographical
		hash function (with some secret key) on these IDs at the server and the client and to compare these hash values
	*/
	boolean isEqual;
	for (int i= 0;i<numberUserIDs;++i) {
		isEqual = true;
		for (int j=0;j<userIdLength;j++) {
			isEqual = isEqual & (userID[j] == buffer[posFirstUserID+4*i+j]);
		}
		if (isEqual) {
			// just for debugging
//			System.out.println("address number: "+(i+1));
			userPreferences.addSubscribedService(serviceName, serviceAlias);
			break;
		}
	}
}


}