// * ======================================================================
// * *                                                                    *
// * *  Copyright (c) 1996, University of Salzburg.                       *
// * *  All rights reserved.                                              *
// * *                                                                    *
// * ======================================================================
// * *                                                                    *  
// * *  DataCast Version 1.8                                              *
// * *                                                                    *
// * *  Author:  Manfred Hackel                                           *
// * *  File  :  Recorder.java                                            *
// * *  Date  :  31.01.1996                                               *
// * *                                                                    *
// * *  Description:                                                      *
// * *                                                                    *
// * *                                                                    *
// * *  Revision-History:                                                 *
// * *     1.7   Manfred Hackel   initial coding                          *
// * *     1.7R  Hilmar Linder    added RRMP support for Windows 95/NT    *
// * *     1.8   Hilmar Linder    check the OS version for RRMP support   *
// * *                            added the options -rf,-ru for RRMP      *
// * *                            added this header                       *
// * *                                                                    *
// * ======================================================================

import java.util.*;
import java.awt.*;
import java.io.*;

public class Recorder {
/* 	component which records statistics about the received datagrams and occuring errors;
	 if attribute 'toFile' is true, output is written to a logfile, if 'toScreen' is true, output is displayed on screen
*/

// instance variables:
	final int datagramLength = 1024;
   	private boolean verbose = true;
	private boolean toScreen = true;				// if 'toScreen' is true all output is written to the screen (stdout)
	private boolean toFile = true;				// if 'toFile' is true all output is written to a log file
	private String logFileName = new String();
	private BufferedOutputStream logOutputStream = null;	// buffered output stream associated with the logfile;
	private static byte[] buffer = new byte[256];			// buffer for writing Strings to byte[], necessary for BufferedOutputStream.write(,,)
	private TextArea outputMessageArea;			// TextArea in the GUI where to display error messages and status report
	private TextField outputStatusLine;				// TextField in the GUI where to display the currently received docuement
	private StatusPanel statusPanel;				// Panel in the GUI that displays the progress of receiving
	private boolean detailedMessages = true; 			// print detailed error messages
	private boolean showStatusLine = true; 			// show status line in GUI
	private boolean showStatusPanel = true;			 // show status panel in GUI
	private String previousServiceName = new String("other");		// name of the service to which the previous Document belonged

	// for statistics:
	private long numberReceivedDocuments = 0;
		// number of documents of which (at least) the first segment has been received
	private long numberRejectedDocuments = 0;
		// number of documents that did not meet the user prefences and thus were not accepted (not stored)
	private long numberStoredDocuments = 0;
		// number of documents that have been accepted and stored successfully
	private long numberReceivedDatagrams = 0;
		// number of datagrams (segments) that have arrived at the receiver
	private long numberRequiredDatagrams = 0;
		// number of datagrams (segments) that should arrive according to the information in the first segments of accepted documents
	private long numberCorrectDatagrams = 0;	
		// number of received datagrams (segments) that belong to  accepted documents; datagrams that belong to other documents are considered 'incorrect' but are counted with 'numberReceivedDatagrams'
	private Date startTime,endTime;			// begin and end of transmission

	// state of object:
	private boolean corruptDocument = false;	
		// for an accepted document of which segments are missing, this variable is set to 'true'
	private boolean truncatedDocument = false;	
		// if the receiver receives a document of which the first segment is missing, this variable is set to true
	private boolean rejectedDocument = false;
		// if the receiver receives a document which does not meet the user preferences, this variable is set to 'true'
	private boolean documentFinished = true;	
		// if a document has been stored, this variable is set to 'true'
	private boolean storageError = false;
		// has a storage error occured during document processing

	// GUI Strings
	private String	txtErrorWritingFile = new String("Error writing to file");
	private String	txtErrorOpeningFile = new String("Error opening file");
	private String	txtErrorDeletingFile = new String("Error deleting file");
	private String	txtAnnounce1 = new String("is arriving");
	private String	txtAnnounce2 = new String("files expected"); 
	private String	txtReceiverError = new String("Fatal error in receiver");
	private String	txtProtocolError = new String("Incompatible protocol versions");
	private String	txtErrorCreatingDirectory = new String("error creating directory");
	private String	txtDiskFull = new String("disk space designated for DataCast is full");

//	********* Recorder ***********
public Recorder(UserPreferences userPreferences, String language, TextArea messageArea, TextField statusLine, StatusPanel statusPanel) {
/*	initialises the Recorder:
	get values for 'toFile' and 'toScreen' and the name of the log file from UserPreferences
	delete the old log file (if it exists)
*/
	outputMessageArea = messageArea;
	outputStatusLine = statusLine;
	this.statusPanel = statusPanel;

	// read GUI txts from file Recorder***.txt
	if (language.compareTo("English")!= 0) {
		String textFileName = new String("Recorder"+language+".txt");
		File f = new File(textFileName);
		if (f.exists()) {
			DataInputStream d = null; 
			try {
				d = new DataInputStream(new FileInputStream(f));
				txtErrorWritingFile = d.readLine();
				txtErrorOpeningFile = d.readLine();
				txtErrorDeletingFile = d.readLine();
				txtAnnounce1 = d.readLine();
				txtAnnounce2 = d.readLine();
				txtReceiverError = d.readLine();
				txtProtocolError = d.readLine();
				txtErrorCreatingDirectory = d.readLine();
				txtDiskFull = d.readLine();
			}
			catch (Exception e) {
				messageArea.appendText(txtErrorOpeningFile+" "+textFileName+"\n");
			}
			finally {
				if (d != null ) {
					try { d.close(); }
					catch ( IOException ioe ) {}
				}
			}
		} 
		else {
			// else default language: English
			messageArea.appendText(txtErrorOpeningFile+" "+textFileName+"\n");
		}
	}

	// reset counters:	
	numberReceivedDocuments = 0;
	numberRejectedDocuments = 0;
	numberStoredDocuments = 0;
	numberReceivedDatagrams = 0;
	numberRequiredDatagrams = 0;
	numberCorrectDatagrams = 0;	

	// go to state <wait for first DG>
	corruptDocument = false;	
	truncatedDocument = false;	
	rejectedDocument = false;
	documentFinished = true;	
	storageError = false;
  	 toScreen = userPreferences.displayOnScreen();
	toFile = userPreferences.recordToFile();
	verbose = DataCast.verbose;
	detailedMessages = userPreferences.detailedErrorMessages();
	showStatusLine = userPreferences.showStatusLine();
	showStatusPanel = userPreferences.showStatusPanel();

	if (toFile) {
		logFileName = userPreferences.getLogFileName();
		try {
			File oldFile = new File(logFileName);
			if (oldFile.exists()) {
				if (!oldFile.delete()) {
					toFile = false;
					writeMessage(txtErrorDeletingFile+" "+logFileName);
				}
			}
			logOutputStream = new BufferedOutputStream(new FileOutputStream(logFileName));
			Date date = new Date();
			println(date.toString());
		} 
		catch (Exception e) {
			toFile = false;
			if (detailedMessages) {
				writeMessage(txtErrorOpeningFile+" "+logFileName+"("+e.toString()+")");
			}
			else {
				writeMessage(txtErrorOpeningFile+" "+logFileName);
			}
		}
	}
}

//	********* println ***********
private void print(String s) {
	// prints the argument to the log file (if 'toFile' is true) and the screen/stdout (if 'toScreen' is true)
	if (toFile) {
		try {
			s.getBytes(0,s.length(),buffer,0);
			logOutputStream.write(buffer,0,s.length());
		} catch (Exception e) {
			toFile = false;
			if (detailedMessages) {
				writeMessage(txtErrorWritingFile+" "+logFileName+"("+e.toString()+")");
//				System.out.println("error while writing "+s);
			}
			else {
				writeMessage(txtErrorWritingFile+" "+logFileName);
			}
		}
	}
	if (toScreen) {
		System.out.print(s);
	}
}

//	********* print ***********
public void println(String s) {
/*	prints the argument plus a line feed and a carriage return 
	to the log file (if 'toFile' is true) and the screen (if 'toScreen' is true)
*/
	print(s+"\n");
}

//	********* writeMessage ***********
public void writeMessage(String text) {
	// displays text in the GUI textField, on stdout and in the log file if selected from user
	outputMessageArea.appendText(text+"\n");
	println(text);	
}

//	********* expectedNewDocument ***********
public void expectedNewDocument (String fileName, String serviceName, long numberSegments) {
/*	tells the Recorder that the first segment of an expected new Document has arrived with
	fileName as the name of the file in which the document has to be stored,
	serviceName as the name of the service to which the document belongs, and
	numberSegments as the number of segments of which this document consists
*/
	if (numberReceivedDatagrams ==0) {
		startTime = new Date();
	}
	if (numberReceivedDocuments == 0) {	
		// go to state <first document arriving>
		documentFinished = false;
	}
	else {
		// go to state < new document arriving>
		documentFinished = true;
	}
	corruptDocument = false;
	rejectedDocument = false;
	truncatedDocument = false;
	storageError = false;
	numberReceivedDocuments ++;
	numberReceivedDatagrams ++;
	numberRequiredDatagrams = numberRequiredDatagrams + numberSegments;
	numberCorrectDatagrams ++;
	if (verbose)
		System.out.println("\nservice name: "+serviceName+", file: "+fileName);
	//else
	//	println("file: "+fileName);
	
	if (showStatusLine) {
		outputStatusLine.setText(fileName);
	}
	if (showStatusPanel) {
		if (serviceName.compareTo("other") == 0 | serviceName.compareTo(previousServiceName) != 0) {
			statusPanel.reset(0);
			previousServiceName = serviceName;
		}
			else {
			statusPanel.increment();
		}
	}
	if (numberSegments == 1) {
		if (verbose)
		   println("1 segment expected");
	}
	else {
		if (verbose)
		   println(numberSegments+" segments expected");
	}
	if (verbose)
	   print("sequence numbers received: 1");
}

//	********* unexpectedNewDocument ***********
public void unexpectedNewDocument (String fileName, String serviceName, long numberSegments) {
/*	tells the Recorder that the first segment of an unexpected new Document has arrived with
	fileName as the name of the file in which the document has to be stored,
	serviceName as the name of the service to which the document belongs, and
	numberSegments as the number of segments of which this document consists
*/
	if (numberReceivedDatagrams ==0) {
		startTime = new Date();
	}
	numberReceivedDocuments ++;
	numberReceivedDatagrams ++;
	numberRequiredDatagrams = numberRequiredDatagrams + numberSegments;
	numberCorrectDatagrams ++;
	if (!rejectedDocument) {
		if (verbose)
			println (" file deleted.");
	}
	// go to state <new document arriving>
	documentFinished = true;
	corruptDocument = false;
	rejectedDocument = false;
	truncatedDocument = false; 
	storageError = false;
	if (verbose)
		System.out.println("\nunexpected service name: "+serviceName+", file: "+fileName);
	//else
	//	println("file: "+fileName);
	
	if (showStatusLine) {
		outputStatusLine.setText(fileName);
	}
	if (showStatusPanel) {
		if (serviceName.compareTo("other") == 0 | serviceName.compareTo(previousServiceName) != 0) {
			statusPanel.reset(0);
			previousServiceName = serviceName;
		}
		else {
			statusPanel.increment();
		}
	}
	if (numberSegments == 1) {
		if (verbose)
			println("1 segment expected");
	}
	else {
		if (verbose)
			println(numberSegments+" segments expected");
	}
	if (verbose)
		print("sequence numbers received: 1");
}
			
//	********* expectedDatagram ***********
public void expectedDatagram (long seqNr) {
/*	tells the Recorder that a segment has arrived which
	* is not the first segment of a document
	* belongs to the same document as the previously received segment
	* has the next sequence number
	* is not part of a corrupt document (a document where previous segments are missing)
	* is not part of a rejected document (a document that did not meet the user preferences)
*/
	// go to state <wait for next segment> (coming from state <first document arriving> all state variables are already false)
	numberCorrectDatagrams ++;
	numberReceivedDatagrams ++;
	if (verbose)
	   print(","+seqNr);
	adjustStatusLine(seqNr);
}

//	********* segmentsMissing ***********
public void segmentsMissing (String fileName, long seqNr, boolean isNewDocument) {
/*	tells the Recorder that a segment has arrived which
	* is not the first segment of a document
	* does NOT belong to the same document as the previously received segment OR
	   has NOT the next sequence number (i.e. segments are missing)
	* is not part of a corrupt document (a document where previous segments are missing)
	* is not part of a rejected document (a document that did not meet the user preferences)
	if isNewDocument = true, the received datagram belongs to a new document, 
	otherwise it belongs to the same document
*/
	if (numberReceivedDatagrams ==0) {
		startTime = new Date();
	}
	numberReceivedDatagrams ++;
	if (verbose)
		print(","+seqNr+" file deleted");
	else
		print(fileName+" - file deleted\n" );
	
	if (showStatusLine) {
		outputStatusLine.setText("");
	}
	if (isNewDocument) {
		if (verbose)
		   print(" - new document arriving");
		// go to state <document is truncated>	(coming from state <first document arriving> all other state variables are false)
		// 					(coming from state <wait for next segment> all other state variables are false)
		truncatedDocument = true;
		corruptDocument = true; 
		documentFinished = true;
	}
	else {
		numberCorrectDatagrams ++;
		// go to state <document is corrupt> 	(coming from state <first document arriving> all other state variables are false)
		// 					(coming from state <wait for next segment> all other state variables are false)
		corruptDocument = true;
	}
}

//	********* otherDatagram ***********
public void otherDatagram (long seqNr) {
/*	tells the Recorder that a segment has arrived which
	* is not the first segment of a document
	* is part of a corrupt/truncated document (a document where previous segments are missing) OR
 	   is part of a rejected document (a document that did not meet the user preferences)
*/
	if (numberReceivedDatagrams == 0) {
		// coming from state <init>, go to state <useless datagram arriving>
		startTime = new Date();
		truncatedDocument = true;
	}
	else {
		if (storageError) {
			numberCorrectDatagrams ++;
		}
		if (rejectedDocument) {
			// say in state <rejected document>
		}
		else {
			if (truncatedDocument) {
				// coming from state <useless datagrams arriving> or <truncated document> 
				// stay in old state
			}
			else {
				if (corruptDocument) {
					// stay in state <corrupt document>
					numberCorrectDatagrams ++;
				}
				else {
					if (documentFinished) {
						// coming state <wait for next document>
						truncatedDocument = true;
						// new state <document is truncated>
					}
					else {
						// no possible state for this event
					}
				}
			}
		}
	}
	numberReceivedDatagrams ++;
	if (verbose)
	   print(","+seqNr);
	adjustStatusLine(seqNr);
}

//	********* truncatedNewDocument ***********
public void truncatedNewDocument() {
/*	in state <document is corrupt> this event tells the Recorder that a segment has arrived
	which belongs to a new document, but which is not the first segment this new Document;
	the Recorder goes to state <document is truncated> */
//	go to state <document is truncated> (coming from <document is corrupt, corruptDocument is already true)
	truncatedDocument = true;
	documentFinished = true;
}

//	********* documentStored ***********
public void documentStored() {
	// tells the Recorder that a document has been stored successfully
	numberStoredDocuments ++;
	// go to state <wait for next document>	(coming from state <first document arriving> all other state variables are false)
	//					(coming from state <wait for next segment> all other state variables are false)
	//					(coming from state <new document arriving> state variables don't change)
	documentFinished = true;
	if (verbose)
	   println(".");
	if (showStatusLine) {
		outputStatusLine.setText("");
	}
}

//             ********* serviceFinished **************
public void serviceFinished() {
	// tells the Recorder that a service has been received completely
	float percent;
	long numberCorruptDocuments;	
	long milliseconds;
	
	writeMessage("Download completed ("+numberStoredDocuments+" files stored successfully)");
	if (showStatusPanel) {
		statusPanel.reset(0);
	}
	
	endTime = new Date();
	println("\nStatistics: \n");

	println(numberReceivedDatagrams+" datagrams received");
	println(numberRequiredDatagrams+" datagrams should arrive according the header of accepted documents");
	if (numberRequiredDatagrams > 0) {
		percent = 100*numberCorrectDatagrams/numberRequiredDatagrams;
		println(numberCorrectDatagrams+" datagrams of these arrived without errors ("+percent+"%)");
		println((numberRequiredDatagrams-numberCorrectDatagrams)+" datagrams were missing ("+(100-percent)+"%)\n");
	}
	
	println(numberReceivedDocuments+" documents received");
	if (numberReceivedDocuments > 0) {
		percent = 100*numberRejectedDocuments / numberReceivedDocuments;
		println(numberRejectedDocuments+" of these have not been accepted ("+percent+"%)");
		percent = 100*numberStoredDocuments / numberReceivedDocuments;
		println(numberStoredDocuments+" documents have been stored successfully ("+percent+"%)");
		numberCorruptDocuments = numberReceivedDocuments - numberRejectedDocuments - numberStoredDocuments;
		percent = 100*numberCorruptDocuments / numberReceivedDocuments;
		println(numberCorruptDocuments+" documents have not been stored succesfully  ("+percent+"%)\n"); 
	}

	if (numberReceivedDatagrams > 0 & startTime != null) {
		milliseconds = endTime.getTime() - startTime.getTime();
		print((datagramLength*numberReceivedDatagrams) + " Bytes received in "+(milliseconds/1000)+" seconds");
		println(" ("+ (datagramLength*numberReceivedDatagrams/(milliseconds*1024/1000)) +" kBps)");
	}
	// reset counters:	
	numberReceivedDocuments = 0;
	numberRejectedDocuments = 0;
	numberStoredDocuments = 0;
	numberReceivedDatagrams = 0;
	numberRequiredDatagrams = 0;
	numberCorrectDatagrams = 0;	
}

//	********* storageError ***********
public void storageError(String fileName, Exception e) {
	// tells the Recorder that there was an error in storing the document
	// go to state <wait for next document> 	(coming from state <first document arriving> all other state variables are false)
	//					(coming from state <wait for next segment> all other state variables are false)
	//					(coming from state <new document arriving> state variables don't change)
	documentFinished = true;
	if (e.toString().endsWith("disk full")) {
		storageError = true;
		writeMessage(txtErrorWritingFile+" "+fileName+" ("+txtDiskFull+")");
	}
	else {
		if (e.toString().endsWith("directory error")) {
			storageError = true;
			writeMessage(txtErrorWritingFile+" "+fileName+" ("+txtErrorCreatingDirectory+")");
		}
		else {
			if (detailedMessages) {
				writeMessage(txtErrorWritingFile+" "+fileName+" ("+e.toString()+")");
			}
			else {
				writeMessage(txtErrorWritingFile+" "+fileName);
			}
		}
	}
}


//	********* announceService ***********
public void announceService(String serviceName, long numberFiles) {
	// announce new service on GUI
	outputMessageArea.appendText('"'+serviceName+'"'+" "+txtAnnounce1+" ("+numberFiles+" "+txtAnnounce2+")\n");
	if (showStatusPanel) {
		statusPanel.reset(numberFiles);
	}
}

//	********* documentRejected ***********
public void documentRejected(long numberSegments) {
	// tells the Recorder that a document with numberSegments segments has been rejected
	// go to state <document is rejected> 	(coming from <first document arriving> all other state variables are false)
	rejectedDocument = true;
	// 					(coming from <new document arriving> documentFinished was true)
	documentFinished = false;
	numberRejectedDocuments ++;
	numberCorrectDatagrams --;
	numberRequiredDatagrams = numberRequiredDatagrams - numberSegments;
	if (verbose)
	   print(" rejected");
	if (showStatusLine) {
		outputStatusLine.setText("");
	}
}

//	********* receiverError ***********
public void receiverError(Exception e) {
	if (detailedMessages) {
		writeMessage(txtReceiverError+" ("+e.toString()+")");
		System.out.println(txtReceiverError+" ("+e.toString()+")");
	}
	else {
		writeMessage(txtReceiverError);
	}
}

//	********* protocolError ***********
public void protocolError(Exception e) {
	if (detailedMessages) {
		writeMessage(txtProtocolError+" ("+e.toString()+")");
	}
	else {
		writeMessage(txtProtocolError);
	}
}

//	********* terminate ***********
public void terminate() {
	// writes statistics to the log file and/or to the screen and terminates recording
	float percent;
	long numberCorruptDocuments;	
	long milliseconds;
/*
	endTime = new Date();
	println("\nStatistics: \n");

	println(numberReceivedDatagrams+" datagrams received");
	println(numberRequiredDatagrams+" datagrams should arrive according the header of accepted documents");
	if (numberRequiredDatagrams > 0) {
		percent = 100*numberCorrectDatagrams/numberRequiredDatagrams;
		println(numberCorrectDatagrams+" datagrams of these arrived without errors ("+percent+"%)");
		println((numberRequiredDatagrams-numberCorrectDatagrams)+" datagrams were missing ("+(100-percent)+"%)\n");
	}
	
	println(numberReceivedDocuments+" documents received");
	if (numberReceivedDocuments > 0) {
		percent = 100*numberRejectedDocuments / numberReceivedDocuments;
		println(numberRejectedDocuments+" of these have not been accepted ("+percent+"%)");
		percent = 100*numberStoredDocuments / numberReceivedDocuments;
		println(numberStoredDocuments+" documents have been stored successfully ("+percent+"%)");
		numberCorruptDocuments = numberReceivedDocuments - numberRejectedDocuments - numberStoredDocuments;
		percent = 100*numberCorruptDocuments / numberReceivedDocuments;
		println(numberCorruptDocuments+" documents have not been stored succesfully  ("+percent+"%)\n"); 
	}

	if (numberReceivedDatagrams > 0 & startTime != null) {
		milliseconds = endTime.getTime() - startTime.getTime();
		print((datagramLength*numberReceivedDatagrams) + " Bytes received in "+(milliseconds/1000)+" seconds");
		println(" ("+ (datagramLength*numberReceivedDatagrams/(milliseconds*1024/1000)) +" kBps)");
	}
*/
}

//	********* close ***********
public void close () {
	if (logOutputStream != null) {
		try {
			logOutputStream.flush();
		}
		catch(Exception e) {
/*			if (detailedMessages) {
				writeMessage(txtErrorWritingFile+" "+logFileName+" ("+e.toString()+")!");
			}
			else {
				writeMessage(txtErrorWritingFile+" "+logFileName+"!");
			}
*/		}
		finally {
			if (logOutputStream != null ) {
				try { logOutputStream.close(); }
				catch ( IOException ioe ) {}
			}
		}
	}
}

//	********* finalize ***********
protected void finalize() {
	close();
}

//	********* adjustStatusLine ***********
private void adjustStatusLine(long seqNr) {
// ad '.' to the filename in the status line; every 500th datagram reset the status line
	if (showStatusLine) {
		String t = new String(outputStatusLine.getText());
		if ((seqNr % 300) == 0) {
			int k = t.indexOf("....");
			if (k>-1) {
				t=t.substring(0,k);
			}
		}
		if ((seqNr % 10) == 0) {
			outputStatusLine.setText(t+".");
		}
	}
}


}