// * ======================================================================
// * *                                                                    *
// * *  Copyright (c) 1996, University of Salzburg.                       *
// * *  All rights reserved.                                              *
// * *                                                                    *
// * ======================================================================
// * *                                                                    *  
// * *  DataCast Version 1.8                                              *
// * *                                                                    *
// * *  Author:  Manfred Hackel                                           *
// * *  File  :  Repository.java                                          *
// * *  Date  :  31.01.1996                                               *
// * *                                                                    *
// * *  Description:                                                      *
// * *                                                                    *
// * *                                                                    *
// * *  Revision-History:                                                 *
// * *     1.7   Manfred Hackel   initial coding                          *
// * *     1.7R  Hilmar Linder    added RRMP support for Windows 95/NT    *
// * *     1.8   Hilmar Linder    check the OS version for RRMP support   *
// * *                            added the options -rf,-ru for RRMP      *
// * *                            added this header                       *
// * *                                                                    *
// * ======================================================================

import java.util.*;
import java.io.*;

public class Repository {
/* 	manages an HTML-file which contains links to the entry pages of the delivered Web documents
	and the repository of all received files
*/
	// instance variables:
	final String IndexHeader = "<HTML>\n<HEAD>\n<TITLE>DataCast</TITLE>\n</HEAD>\n<BODY>\n<H1>Internet Data Multicast Service</H1>\nThe following services are available:\n<HR>\n<UL>\n";
	final String IndexTrailer = "</UL>\n<HR>\n</BODY>\n</HTML>";
	private long freeDiskSpace;		// # of Bytes free for user data
	private long fileExpiryTime;		// milli-seconds
	private Recorder recorder;
	private UserPreferences userPreferences;
	private Hashtable linkedServices = new Hashtable();	// services to which Index.html provides hyperlinks
	private Hashtable storedFiles = null;		// Hashtable of stored files with their date of storage
		// unfortunately File.lastModified() can not be used as an absolute time, therefore I have to store the dates 

	private boolean detailedMessages = true; // print detailed error messages

	// GUI strings
	private String		txtErrorWritingFile = new String("Error writing to file");
	private String		txtErrorOpeningFile = new String("Error opening file");
	private String		txtErrorReadingFile = new String("Error reading file");
	private String		txtFreeSpace = new String("bytes on disk are designated for incoming files");
	private String		txtDeleteFiles = new String("deleting old files ...");
	private String		txtCheckDisk = new String("checking free disk space ...");

//	******** Repository ************
public Repository(UserPreferences userPreferences, Recorder recorder, String language) {
	this.recorder = recorder;
	this.userPreferences = userPreferences;
	String serviceName;
	String fileName;
	Date date;
	long serviceExpiryTime = 3600000*userPreferences.getServiceExpiryTime();
	fileExpiryTime = 3600000*userPreferences.getFileExpiryTime();
	linkedServices.clear();

	// read GUI txts from file Repository***.txt
	String textFileName = new String("Repository"+language+".txt");
	File f;
	if (language.compareTo("English") != 0) {
		f  = new File(textFileName);
		if (f.exists()) {
			DataInputStream d = null;
			try {
				d = new DataInputStream(new FileInputStream(f));
				txtErrorWritingFile = d.readLine();
				txtErrorOpeningFile = d.readLine();
				txtErrorReadingFile = d.readLine();
				txtFreeSpace = d.readLine();
				txtDeleteFiles = d.readLine();
				txtCheckDisk = d.readLine();
			}
			catch (Exception e) {
				recorder.writeMessage(txtErrorReadingFile+" "+textFileName);	
			}
			finally {
				if (d != null ) {
					try { d.close(); }
					catch ( IOException ioe ) {}
				}
			}
		}
		else { // else default language: English
			recorder.writeMessage(txtErrorOpeningFile+" "+textFileName);	
		}
	}

	// Hashtable of storedFiles:
	recorder.writeMessage(txtDeleteFiles);
	storedFiles = new Hashtable();
	RandomAccessFile raf = null;
	fileName = new String();
	Date fileDate = new Date();
	try {
		raf = new RandomAccessFile("storedFiles.txt","r");
		if (raf.length() > 0) {
			StoredFile storedFile;
			try {
				while (true) {
					fileName = raf.readLine();
					fileDate = new Date(raf.readLine());
					storedFile = new StoredFile(fileName,fileDate);
					storedFiles.put(fileName,storedFile);
				}
			}
			catch (IOException e) {} // EOF
		}
	}
	catch (IOException e) {
		recorder.writeMessage(txtErrorReadingFile+" storedFiles.txt ("+e.toString()+")");	
	}
	catch(Exception e) {}
	finally {
		if (raf != null ) {
			try { raf.close(); }
			catch ( IOException ioe ) {}
		}
	}

	// delete stored files that are older than fileExpiryTime:
	long nowLong = (new Date()).getTime();
	StoredFile storedFile;
	for (Enumeration s = storedFiles.elements(); s.hasMoreElements();) {
		storedFile = (StoredFile) (s.nextElement());
		try {
			f = new File(storedFile.name.replace('/',File.separatorChar));
			if (f.exists()) {
				if (nowLong - storedFile.date.getTime() > fileExpiryTime) {
					f.delete();
					storedFiles.remove(storedFile.name);
				}
			}
			else {
				storedFiles.remove(storedFile.name);
			}
		}
		catch (Exception e) {
			System.out.println("<Repository.init>: "+e.toString());
		}
	}

	// delete empty subdirectories:
	try {
		File inputFile = new File(".");
		if (inputFile.exists()) {
			if (inputFile.isDirectory()) {
				String fileList[] = new String[5];
				fileList = inputFile.list();
				for (int i=0;i<fileList.length;++i) {
					f = new File(fileList[i]);
					if (f.isDirectory()) {				
						deleteDirectory(fileList[i]);
					}
				}
			}
		}
	}
	catch(Exception e) {
		recorder.writeMessage("<Repository.init>: "+e.toString());
	}

	// remove old entries from Index.html:
	RandomAccessFile indexFile = null;
	RandomAccessFile backUpFile = null;
	try {
		// delete old index file if it exists
		File oldFile = new File("Index.html");
		if (!oldFile.exists() | (oldFile.exists() & oldFile.delete())) {
			// generate new index file
			indexFile = new RandomAccessFile("Index.html","rw");	// HTML-file containing the table of content 
			indexFile.writeBytes(IndexHeader);
			// add services listed in services.txt  which have not expired up to now
			File oldBackUpFile = new File("services.txt");
			if (oldBackUpFile.exists()) {
				backUpFile = new RandomAccessFile("services.txt","rw");
				backUpFile.seek(0);
				Date now = new Date();
				try {
					while (true) {
						serviceName = backUpFile.readLine();
						fileName = backUpFile.readLine();
						date = new Date(backUpFile.readLine());
						Service service = new Service(serviceName,fileName,date);
						if (now.getTime() - date.getTime() < serviceExpiryTime) {
							linkedServices.put(service.aliasName(),service);
							try {
						 		indexFile.writeBytes("<LI><A HREF=\""+fileName+"\">"+serviceName+"</A>\n");
							} catch (Exception e) {
								if (detailedMessages) {
									recorder.writeMessage(txtErrorWritingFile+" Index.html ("+e.toString()+")");
								}
								else {
									recorder.writeMessage(txtErrorWritingFile+" Index.html");
								}
								throw e;
							}
						}
					}
				} catch (Exception e) { // end of file or error in writing to Index.html
				}
				backUpFile.close();
			}
			oldBackUpFile.delete();
			indexFile.writeBytes(IndexTrailer);
			indexFile.close();
		} // else: error in deleting the old Index.html - do nothing
	} catch (Exception e) {
		if (detailedMessages) {
			recorder.writeMessage(txtErrorWritingFile+" Index.html ("+e.toString()+")");
		}
		else {
			recorder.writeMessage(txtErrorWritingFile+" Index.html");
		}

	}
	finally {
		if (indexFile != null ) {
			try { indexFile.close(); }
			catch ( IOException ioe ) {}
		}
		if (backUpFile != null ) {
			try { backUpFile.close(); }
			catch ( IOException ioe ) {}
		}
	}

}


//	******** close ************
public void close() {
	// store linked services to file services.txt
	try {	
		File f = new File("services.txt");
		if (f.exists()) {
			 f.delete();
		}
	}
	catch (Exception e) {}
	RandomAccessFile file = null;
	try {
		file = new RandomAccessFile("services.txt","rw");
		Service service;
		for (Enumeration s = linkedServices.elements(); s.hasMoreElements();) {
			service = (Service) (s.nextElement());
			file.writeBytes(service.aliasName()+"\n");
			file.writeBytes(service.urlString()+"\n");
			file.writeBytes(service.date().toGMTString()+"\n");
		}
		file.close();
	} 
	catch (Exception e) {
		if (detailedMessages) {
			recorder.writeMessage(txtErrorWritingFile+" services.txt ("+e.toString()+")");
		}
		else {
			recorder.writeMessage(txtErrorWritingFile+" services.txt");
		}
	}
	finally {
		if (file != null ) {
			try { file.close(); }
			catch ( IOException ioe ) {}
		}
	}

	// store storedFiles to file storedFiles.txt
	try {	
		File f = new File("storedFiles.txt");
		if (f.exists()) {
			 f.delete();
		}
	}
	catch (Exception e) {}
	file = null;
	try {
		file = new RandomAccessFile("storedFiles.txt","rw");
		StoredFile storedFile;
		for (Enumeration s = storedFiles.elements(); s.hasMoreElements();) {
			storedFile = (StoredFile) (s.nextElement());
			file.writeBytes(storedFile.name+"\n");
			file.writeBytes(storedFile.date.toGMTString()+"\n");
		}
		file.close();
	} 
	catch (Exception e) {
		if (detailedMessages) {
			recorder.writeMessage(txtErrorWritingFile+" storedFiles.txt ("+e.toString()+")");
		}
		else {
			recorder.writeMessage(txtErrorWritingFile+" storedFiles.txt");
		}
	}
	finally {
		if (file != null ) {
			try { file.close(); }
			catch ( IOException ioe ) {}
		}
	}

}

//	******** checkDiskSpace **********
public void checkDiskUsage() {
	// calculate number of bytes occupied from files in subdirectories of the working directory:
	recorder.writeMessage(txtCheckDisk);
	long diskUsage = 0; 	// sum of file lengths of all files in subdirectories of current directory
	try {
		File inputFile = new File(".");
		if (inputFile.exists()) {
			if (inputFile.isDirectory()) {
				String fileList[] = new String[5];
				fileList = inputFile.list();
				for (int i=0;i<fileList.length;++i) {
					File f = new File(fileList[i]);
					if (f.isDirectory()) {				
						diskUsage = diskUsage + totalFileLength(fileList[i]);
					}
				}
			}
		}
	}
	catch(Exception e) {
		recorder.writeMessage("<Repository.init>: "+e.toString());
	}
	freeDiskSpace = 1024*1024*userPreferences.getMaxDiskSpace()-diskUsage;
	if (freeDiskSpace < 0) {
		freeDiskSpace = 0;
	}
	recorder.writeMessage(freeDiskSpace+" "+txtFreeSpace);
}

//	******** finalize ************
protected void finalize() {
	close();
}

//	******** addFile **********
public void addFile(String fileName) {
	//adds this fileName + its date to the hashtable of stored files
	storedFiles.put(fileName,new StoredFile(fileName,new Date()));
}

//	******** addService ************
public void addService(Service service) {
	// adds to file Index.html a hyperlink to the frontpage of this service 
	Service oldService = null; 
	oldService = (Service) linkedServices.get(service.aliasName());

//	if (linkedServices.containsKey(service.aliasName())) {
	if (oldService != null && oldService.urlString().compareTo(service.urlString()) == 0) {
		linkedServices.remove(linkedServices.get(service.aliasName()));
		linkedServices.put(service.aliasName(), service);
	}
	else {
		if (oldService != null) { // there is an entry with the same service name, but different URL (i.e. edition of yesterday)
			service.setAlias(service.aliasName()+" ("+service.urlString()+")");
		}
		if (linkedServices.containsKey(service.aliasName())) {
			linkedServices.remove(linkedServices.get(service.aliasName()));
			linkedServices.put(service.aliasName(), service);
		}
		else {
			linkedServices.put(service.aliasName(), service);
			RandomAccessFile indexFile = null;
			try {
				// the file is opened at every write operation, to let the Web browser access it during transmission
				indexFile = new RandomAccessFile("Index.html","rw");
				try {
					indexFile.skipBytes((int) indexFile.length() - IndexTrailer.length()); 
					indexFile.writeBytes("<LI><A HREF=\""+service.urlString()+"\">"+service.aliasName()+"</A>\n");
					indexFile.writeBytes(IndexTrailer);
					indexFile.close();
				} catch(Exception e) {
					if (detailedMessages) {
						recorder.writeMessage(txtErrorWritingFile+" Index.html ("+e.toString()+")");
					}
					else {
						recorder.writeMessage(txtErrorWritingFile+" Index.html");
					}
				}
				finally {
					if (indexFile != null ) {
						try { indexFile.close(); }
						catch ( IOException ioe ) {}
					}
				}
			} catch (Exception e) {
				if (detailedMessages) {
					recorder.writeMessage(txtErrorOpeningFile+" Index.html ("+e.toString()+")");
				}
				else {
					recorder.writeMessage(txtErrorOpeningFile+" Index.html");
				}
			}
			finally {
				if (indexFile != null ) {
					try { indexFile.close(); }
					catch ( IOException ioe ) {}
				}
			}
		}
	}
}

//	******** deleteDirectory ************
private void deleteDirectory (String fileName) {
//	deletes this directory if it is empty
//	System.out.println(fileName);
	File file = new File(fileName);
	if (file.isDirectory()) {
		String fileList[] = new String[5];
		fileList = file.list();
		for (int i=0;i<fileList.length;++i) {
			deleteDirectory(fileName+file.separator+fileList[i]);
		}
		if (file.list().length == 0) {
			boolean b = file.delete();
			// seems to work not on directories; I don't know how to delete directories; there is no other method
		}
	}
}

//	******** isEnoughFreeDiskSpace ************
public boolean isEnoughFreeDiskSpace(long fileLength) {
	return (fileLength < freeDiskSpace);
}

//	******** decrementFreeDiskSpace ***********
public void decrementFreeDiskSpace (long amount) throws Exception {
//	decrements the free (reserved) disk space by amount Bytes
	if (amount > freeDiskSpace) {
		freeDiskSpace = 0;
		Exception e = new Exception("DataCast disk storage capacity surpassed");
		throw e;
	}
	else {
		freeDiskSpace = freeDiskSpace-amount;
	}
}

//	******** totalFileLength ************
private long totalFileLength (String fileName) throws Exception {
	// returns to the total length of all files in this directory and all its subdirectories
	File inputFile = new File(fileName);
	if (inputFile.exists()) {
		if (inputFile.isDirectory()) {
			long fileLength = 0;
			String fileList[] = new String[5];
			fileList = inputFile.list();
			for (int i=0;i<fileList.length;++i) {
				fileLength = fileLength + totalFileLength(fileName+inputFile.separator+fileList[i]);
			}
			return fileLength;
		}
		else { // not a directory
			return inputFile.length();
		}
	}
	else {
		return 0;
	}
}
 
}

//	******** StoredFile ************
class StoredFile extends Object {
	String name;
	Date date;

	public StoredFile(String name, Date date) {
		this.name = name;
		this.date = date;
	}
}


