// * ======================================================================
// * *                                                                    *
// * *  Copyright (c) 1996, University of Salzburg.                       *
// * *  All rights reserved.                                              *
// * *                                                                    *
// * ======================================================================
// * *                                                                    *  
// * *  DataCast Version 1.8                                              *
// * *                                                                    *
// * *  Author:  Manfred Hackel                                           *
// * *  File  :  UserProferences.java                                     *
// * *  Date  :  31.01.1996                                               *
// * *                                                                    *
// * *  Description:                                                      *
// * *                                                                    *
// * *                                                                    *
// * *  Revision-History:                                                 *
// * *     1.7   Manfred Hackel   initial coding                          *
// * *     1.7R  Hilmar Linder    added RRMP support for Windows 95/NT    *
// * *     1.8   Hilmar Linder    check the OS version for RRMP support   *
// * *                            added the options -rf,-ru for RRMP      *
// * *                            added this header                       *
// * *                                                                    *
// * ======================================================================

import java.util.*;
import java.io.*;
import java.awt.TextArea;

public class UserPreferences {
/*	manages the user preferences and program options,
	e.g. which documents should be accepted ?,
	should the processing be recorded in a log file ?
*/

// instance variables:
	private DataCast GUI;			// reference to instance of Graphical User Interface
	private boolean recordToFile = true; 	// should the recorder write its output to a log file?
	private boolean displayOnScreen = false; 	// should the Recorder display its output on the screen (stdout)
	private String logFileName = "logfile.txt";	// name of the log file
	private int maxDiskSpace = 100; 		// maximal disk space in MB used to store received files + DataCast programme files
	private long serviceExpiryTime = 24;	// unit hours!
		/*	entries in the table of content (Index.html) that are older than 'serviceExpiryTime' (unit: hours) will be removed 
			at a new start of DataCast Receiver */
	private long fileExpiryTime = 24; 	// unit hours!
		/*	files in subdirectories of the working directory of DataCast Receiver that are older than 'fileExpiryTime' (unit: hours) 
			will be deleted at a new start of DataCast Receiver */
	private int port = 4440;	// UDP port number where the receiver will wait for incoming data
	private String multicastAddr = "233.1.1.1";   // RRMP multicast address

	private Hashtable subscribedServices = new Hashtable();
		// services that the user has subscribed
	private Vector acceptedServices = new Vector();
		// services that the user is willing to accept (if the document does not belong to any of these services, it will be rejected)
	private Vector acceptedSubjects = new Vector();
		// subjects that the user is willing to accept (if the document has at least one of these subjects in its description - or no subject - it will be accepted)
	private Vector acceptedRegions = new Vector();
		// regions that the user is willing to accept (if the document has at least one of these regions in its description - or no region - it will be accepted)
	private Vector acceptedMediaTypes = new Vector();
		// media types that the user is willing to accept (if the document has at least one of these types (e.g. "Image", "Text") in its description it will be accepted)
	private Vector acceptedLanguages = new Vector();
		// languages that the user is willing to accept (if the document has one of these lanuages in its description - or no language - it will be accepted)
	private boolean detailedErrorMessages = true;
		// should the displayed error message contain as well the java exception string
	private boolean showStatusLine = true;
		// should the status line be displayed in the GUI
	private boolean showStatusPanel = true;
		// should a status panel be displayed in the GUI

//	******** UserPreferences ************
public UserPreferences(DataCast GUI, TextArea messageArea, String txtErrorReadingFile) {
	this.GUI = GUI;
	// read preferences from file Preferences.ini:
	RandomAccessFile preferencesFile = null;
	try {
		preferencesFile = new RandomAccessFile("Preferences.ini","r");
		try {
			String s = new String();
			String option = new String();
			while (true) {
				s = preferencesFile.readLine();
				if (s.length() > 0) {
					if (s.charAt(s.length()-1) < ' ') {
						// i.e. if readLine() didn't skip the CR or LF
						s = s.substring(0,s.length()-1);
					}
				}
				if ( !s.startsWith("rem") & !s.startsWith("REM") & !(s.length() == 0) ) {
					if (s.startsWith("[")) { // begin of a new option
						option = s.substring(1,s.length()-1);
					}
					else { // possible value for this option
						if (option.compareTo("Services") == 0) {
							acceptedServices.addElement(s);
						}
						if (option.compareTo("Subjects") == 0) {
							acceptedSubjects.addElement(s);
						}
						if (option.compareTo("Media Types") == 0) {
							acceptedMediaTypes.addElement(s);
						}
						if (option.compareTo("Regions") == 0) {
							acceptedRegions.addElement(s);
						}
						if (option.compareTo("Languages") == 0) {
							acceptedLanguages.addElement(s);
						}
						if (option.compareTo("View") == 0) {
							if (s.startsWith("detailedErrorMessages")) {
								detailedErrorMessages = s.endsWith("true");
							}
							if (s.startsWith("showStatusLine")) {
								showStatusLine = s.endsWith("true");
							}
							if (s.startsWith("showStatusPanel")) {
								showStatusPanel = s.endsWith("true");
							}
						}
						if (option.compareTo("Storage") == 0) {
							if (s.startsWith("displayOnScreen")) {
								displayOnScreen = s.endsWith("true");
							}
							if (s.startsWith("logFileName")) {
								logFileName = s.substring(12);
								recordToFile = (logFileName.length() > 0);
							}
							if (s.startsWith("maxDiskSpace")) {
								maxDiskSpace = (new Integer(s.substring(13))).intValue();
							}
							if (s.startsWith("serviceExpiryTime")) {
								serviceExpiryTime = (new Integer(s.substring(18))).longValue();
							}
							if (s.startsWith("fileExpiryTime")) {
								fileExpiryTime = (new Integer(s.substring(15))).longValue();
							}
							if (s.startsWith("port")) {
								port = (new Integer(s.substring(5))).intValue();
							}
							if (s.startsWith("multicast")) {
								multicastAddr = s.substring(10,s.length());
							}
						}
					}
				}
			}
		}
		catch (Exception e) {} // end of file
	}
	catch (Exception e) {
		messageArea.appendText(txtErrorReadingFile+" Preferences.ini");
	}	
	finally {
		if (preferencesFile != null ) {
			try { preferencesFile.close(); }
			catch ( IOException ioe ) {}
		}
	}

	// just for debugging:
	System.out.println("\nServices:");
	for (int i = 0; i < acceptedServices.size();++i) {
		System.out.println(acceptedServices.elementAt(i).toString()+" ");
	}
	System.out.println("\nSubjects:");
	for (int i = 0; i < acceptedSubjects.size();++i) {
		System.out.println(acceptedSubjects.elementAt(i).toString()+" ");
	}
	System.out.println("\nLanguages: "); 
	for (int i = 0; i < acceptedLanguages.size();++i) {
		System.out.println(acceptedLanguages.elementAt(i).toString()+" ");
	}
	System.out.println("maxDiskSpace: "+maxDiskSpace);
	System.out.println("serviceExpiryTime: "+serviceExpiryTime);
	System.out.println("fileExpiryTime: "+fileExpiryTime);

}

//	******** close ************
public void close() {
	// store preferences to file Preferences.ini
	// delete old Preferences.ini if it eists
	RandomAccessFile preferencesFile = null;
	try {
		File oldFile = new File("Preferences.ini");
		if (!oldFile.exists() | (oldFile.exists() & oldFile.delete())) {
			preferencesFile = new RandomAccessFile("Preferences.ini","rw");
			preferencesFile.writeBytes("[Services]\n");
			for (int i = 0; i< acceptedServices.size();++i) {
				preferencesFile.writeBytes( ((String) (acceptedServices.elementAt(i))) +"\n");
			}
			preferencesFile.writeBytes("\n");
			preferencesFile.writeBytes("[Subjects]\n");
			for (int i = 0; i< acceptedSubjects.size();++i) {
				preferencesFile.writeBytes( ((String) (acceptedSubjects.elementAt(i))) +"\n");
			}
			preferencesFile.writeBytes("\n");
			preferencesFile.writeBytes("[Regions]\n");
			for (int i = 0; i< acceptedRegions.size();++i) {
				preferencesFile.writeBytes( ((String) (acceptedRegions.elementAt(i))) +"\n");
			}
			preferencesFile.writeBytes("\n");
			preferencesFile.writeBytes("[Media Types]\n");
			for (int i = 0; i< acceptedMediaTypes.size();++i) {
				preferencesFile.writeBytes( ((String) (acceptedMediaTypes.elementAt(i))) +"\n");
			}
			preferencesFile.writeBytes("\n");
			preferencesFile.writeBytes("[Languages]\n");
			for (int i = 0; i< acceptedLanguages.size();++i) {
				preferencesFile.writeBytes( ((String) (acceptedLanguages.elementAt(i))) +"\n");
			}
			preferencesFile.writeBytes("\n");
			preferencesFile.writeBytes("[View]\n");
				preferencesFile.writeBytes("detailedErrorMessages="+detailedErrorMessages+"\n");
				preferencesFile.writeBytes("showStatusLine="+showStatusLine+"\n");
				preferencesFile.writeBytes("showStatusPanel="+showStatusPanel+"\n");
			preferencesFile.writeBytes("\n");
			preferencesFile.writeBytes("[Storage]\n");
				preferencesFile.writeBytes("displayOnScreen="+displayOnScreen+"\n");
				preferencesFile.writeBytes("logFileName="+logFileName+"\n");
				preferencesFile.writeBytes("maxDiskSpace="+maxDiskSpace+"\n");
				preferencesFile.writeBytes("serviceExpiryTime="+serviceExpiryTime+"\n");
				preferencesFile.writeBytes("fileExpiryTime="+fileExpiryTime+"\n");
				preferencesFile.writeBytes("port="+port+"\n");
				preferencesFile.writeBytes("multicast="+multicastAddr+"\n");
			preferencesFile.writeBytes("\n");
		}
	} catch (Exception e) {
		System.out.println("error writing to file Preferences.ini");
	}
	finally {
		if (preferencesFile != null ) {
			try { preferencesFile.close(); }
			catch ( IOException ioe ) {}
		}
	}
}

//	******** finalize ************
protected void finalize() {
	close();
}

//	******** isAccepted ************
public boolean isAccepted(Document document) {
	// does the document fullfil the user's preferences?
/*	a document with an absolute path name or which should be located outside of a subdirectory of the current directory will not be accepted
	a document with service Name "System" and extention .txt will be acccepted 
	a document with service Name different from "System" and not in the list of accepted or subscribed services will not be accepted
	of all documents with accepted service name:
		the frontpage of the document will be accepted
		a document with a language name not in the list of accepted languages will not be accepted
		of all documents with an accepted language or no language:
			a document with subjects of which not any subject is in the list of accepted subjects will not be accepted
			of all documents with an accepted subject or no subjects:
				a document with regions of which not any region is in the list of accepted regions will not be accepted
				a document with an accepted region or no region will be accepted
*/
	String f = document.fileName();
	if ( 	f.startsWith("/") | 			// absolute path name
		f.startsWith(".") |			// may be used to circumvent other path name checks
		f.indexOf("..") > -1 |		// descending path name
		f.indexOf(':') > -1 ) {		// absolute path name 
//			System.out.println("illegal path");
			return false;
	}
	
	if (document.serviceName().compareTo("System")== 0) { // used for upgrading DataCast Receiver software
		if (document.fileName().endsWith(".txt")) {
			return true;
		}
		else {
			/* accepting .class files to be replaced automatically will result in a security hole
			    since this would allow third parties to replace current files with trojan horses */
//			System.out.println("illegal System file");
			return false;
		}
	}


	if (f.indexOf("/") < 0) {
		// don't allow writing into current directory (only subdirectories are accepted)
//		System.out.println("illegal directory");
		return false;
	}

	// not "System", check if serviceName is accepted:
	if (document.isFree()) {
		if (!acceptedServices.contains(document.serviceName())) {
//			System.out.println("free service rejected");
			return false;
		}
	}
	else {
		if (!subscribedServices.containsKey(document.serviceName())) {
//			System.out.println("subscribed service rejected");
			return false;
		}
	}

	// all front pages will be accepted:
	if (document.isFrontpage()) {
		return true;
	}

	// check for media type:
	if (!acceptedMediaTypes.contains(document.mediaType())) {
//		System.out.println("media type rejected");
		return false;
	}

	// check for language:
	if (	(document.language().length() > 0) &
		(!acceptedLanguages.contains(document.language()))	) {
//		System.out.println("language rejected");
		return false;
	}

	// check for subjects:	
	boolean found = false;
	Vector values = new Vector();
	values = document.subjects();
	if (values.size() == 0) {
		found = true;
	}
	else {
		for (int i = 0; i<values.size();++i) {
			if (acceptedSubjects.contains(values.elementAt(i))) {
				found = true;
				break;
			}
		}
	}

	// check for regions:
	if (found) { 
		found = false;
		values = document.regions();
		if (values.size() == 0) {
			found = true;
		}
		else {
			for (int i = 0; i<values.size();++i) {
				if (acceptedRegions.contains(values.elementAt(i))) {
					found = true;
					break;
				}
			}
		}
	}

	return found;
}

//	******** recordToFile ************
public boolean recordToFile() {
	//	should the recorder write its output to a log file?
	return recordToFile;
}

//	******** displayOnScreen ************
public boolean displayOnScreen() {
	//	should the Recorder display its output on the screen (stdout)
	return displayOnScreen;
}

//	******** enableRecordToFile ************
public void enableRecordToFile() {
	recordToFile = true;
}

//	******** disableRecordToFile ************
public void disableRecordToFile() {
	recordToFile = false;
}

//	******** enableDisplayOnScreen ************
public void enableDisplayOnScreen() {
	displayOnScreen = true;
}

//	******** disableDisplayOnScreen ************
public void disableDisplayOnScreen() {
	displayOnScreen = false;
}

//	******** getLogFileName ************
public String getLogFileName() {
	//	return the name of the log file to which the Recorder should write his output
	if (recordToFile) {
		return logFileName;
	}
	else {
		return null;
	}
}

//	******** setLogFileName ************
public void setLogFileName(String newLogFileName) {
	// to do
}

//	******** getServiceExpiryTime ************
public long getServiceExpiryTime() {
	return serviceExpiryTime;
}

//	******** setServiceExpiryTime ************
public void setServiceExpiryTime(int hours) throws Exception{
	if (hours >= 0) {
		serviceExpiryTime = hours;
	}
	else {
		Exception e = new Exception("illegal value");
		throw e;
	}
}

//	******** getFileExpiryTime ************
public long getFileExpiryTime() {
	return fileExpiryTime;
}

//	******** setFileExpiryTime ************
public void setFileExpiryTime(int hours) throws Exception{
	if (hours >= 0) {
		fileExpiryTime = hours;
	}
	else {
		Exception e = new Exception("illegal value");
		throw e;
	}
}

//	******** setOptionValues ************
public void setOptionValues(String option, Vector acceptedValues) {
/*	sets the list of accepted values for the given option, 
	e.g. if option = "Subjects" and acceptedValues = <"Politics", "Economy", "Science"> the attribute
	acceptedSubjects will be set to this list of accepted values */

/*	System.out.println("option "+option);
	for (int i = 0; i < acceptedValues.size();++i) {
		System.out.println(acceptedValues.elementAt(i).toString());
	}
*/
	String s;
	if (option.compareTo("Subjects") == 0) {
		acceptedSubjects = acceptedValues;
	} 
	if (option.compareTo("Services") == 0) {
		acceptedServices = acceptedValues;
/*		acceptedServices.removeAllElements();
		for (int i = 0; i< acceptedValues.size();++i) {
			s = new String(acceptedValues.elementAt(i).toString());
			System.out.println(s);
			acceptedServices.addElement(s);
		}
*/
/*		System.out.println("\nServices:");
		for (int i = 0; i < acceptedServices.size();++i) {
			System.out.print(acceptedServices.elementAt(i).toString()+" ");
		}
*/	}
	if (option.compareTo("Regions") == 0) {
		acceptedRegions = acceptedValues;
	}
	if (option.compareTo("Media Types") == 0) {
		acceptedMediaTypes = acceptedValues;
	}
	if (option.compareTo("Languages") == 0) {
		acceptedLanguages = acceptedValues;
	}
}

//	******** getAcceptedValues ************
public Vector getAcceptedValues(String option) {
	if (option.compareTo("Subjects") == 0) {
		return acceptedSubjects;
	} 
	if (option.compareTo("Services") == 0) {
		return acceptedServices;
	}	
	if (option.compareTo("Regions") == 0) {
		return acceptedRegions;
	}
	if (option.compareTo("Media Types") == 0) {
		return acceptedMediaTypes;
	}
	if (option.compareTo("Languages") == 0) {
		return acceptedLanguages;
	}
	return (new Vector());
}

//	******** detailedErrorMessages ************
public boolean detailedErrorMessages() {
	// should the displayed error message contain as well the java exception string
	return detailedErrorMessages;
}

//	******** setDetailedErrorMessages ************
public void setDetailedErrorMessages(boolean b) {
	detailedErrorMessages = b;
}


//	******** showStatusPanel ************
public boolean showStatusPanel() {
	// should a status panel be displayed in the GUI
	return showStatusPanel;
}

//	******** setShowStatusPanel ************
public void setShowStatusPanel(boolean b) {
	showStatusPanel = b;
	if (b) {
		GUI.showStatusPanel();
	}
	else {
		GUI.hideStatusPanel();
	}
}

//	******** showStatusLine ************
public boolean showStatusLine() {
	// should the status line be displayed in the GUI
	return showStatusLine;
}

//	******** setShowStatusLine ************
public void setShowStatusLine(boolean b) {
	showStatusLine = b;
	if (b) {
		GUI.showStatusLine();
	}
	else {
		GUI.hideStatusLine();
	}
}

//	******** addSubscribedService ************
public void addSubscribedService(String name, String alias) {
/*	add service with this service name and this service alias name (see next operation) to the table
	of subscribed services
*/
	Service service = new Service(name, null, null, alias);
	try {
		subscribedServices.put(name,service);
	}
	catch(Exception e) {
		System.err.println(e.toString());
	}
}

//	******** getAliasName ************
public String getAliasName(String serviceName) {
/*	return the alias name (real name) of a service that is only available on subscription
	in this case service name is a random string set from the DataCast sender to prevent the Receiver 
	of non-subcribers from storing these file
*/
	Service s = (Service) subscribedServices.get(serviceName);
	if (s == null) {
		return "";
	}
	else {
		return s.aliasName();
	}
}

//	******** getMaxDiskSpace **********
public int getMaxDiskSpace() {
	return maxDiskSpace;
}

//	******** setMaxDiskSpace **********
public void setMaxDiskSpace(int value) throws Exception {

	// to do: check if there is enough space free on hard disk!
	if (value >= 0) {
		maxDiskSpace = value;
	}
	else {
		Exception e = new Exception("illegal value");
		throw e;
	}
}

//	******** getPortNumber **********
public int getPortNumber() {
	return port;
}

//	******** setPortNumber **********
public void setPortNumber(int value) throws Exception {

	if (value >= 1000 & value < 10000) {
		port = value;
	}
	else {
		Exception e = new Exception("illegal value");
		throw e;
	}
}

//	******** getMulticastAddr **********
public String getMulticastAddr() {
	return multicastAddr;
}

//	******** setMulticastAddr **********
public void setMulticastAddr(String addr) {
	multicastAddr = addr;
}


}